from pathlib import Path
from typing import Optional, Text, Union

from ruamel.yaml.error import (
    MarkedYAMLError,
    MarkedYAMLWarning,
    MarkedYAMLFutureWarning,
)


class FileNotFoundException(FileNotFoundError):
    """Raised when a file, expected to exist, doesn't exist."""


class FileIOException(Exception):
    """Raised if there is an error while doing file IO."""


class YamlSyntaxException(Exception):
    """Raised when a YAML file can not be parsed properly due to a syntax error."""

    def __init__(
        self,
        filename: Optional[Union[Text, Path]] = None,
        underlying_yaml_exception: Optional[Exception] = None,
    ) -> None:
        """Represents the exception constructor."""
        self.filename = filename

        self.underlying_yaml_exception = underlying_yaml_exception

    def __str__(self) -> Text:
        if self.filename:
            exception_text = f"Failed to read '{self.filename}'."
        else:
            exception_text = "Failed to read YAML."

        if self.underlying_yaml_exception:
            if isinstance(
                self.underlying_yaml_exception,
                (MarkedYAMLError, MarkedYAMLWarning, MarkedYAMLFutureWarning),
            ):
                self.underlying_yaml_exception.note = None
            if isinstance(
                self.underlying_yaml_exception,
                (MarkedYAMLWarning, MarkedYAMLFutureWarning),
            ):
                self.underlying_yaml_exception.warn = None
            exception_text += f" {self.underlying_yaml_exception}"

        if self.filename:
            exception_text = exception_text.replace(
                'in "<unicode string>"', f'in "{self.filename}"'
            )

        exception_text += (
            "\n\nYou can use https://yamlchecker.com/ to validate the "
            "YAML syntax of your file."
        )
        return exception_text



DEFAULT_SERVER_PORT = 5055
DEFAULT_SANIC_WORKERS = 1
ENV_SANIC_WORKERS = "ACTION_SERVER_SANIC_WORKERS"
DEFAULT_LOG_LEVEL_LIBRARIES = "ERROR"
ENV_LOG_LEVEL_LIBRARIES = "LOG_LEVEL_LIBRARIES"
APPLICATION_ROOT_LOGGER_NAME = "rasa_sdk"
DEFAULT_ENCODING = "utf-8"
YAML_VERSION = (1, 2)
PYTHON_LOGGING_SCHEMA_DOCS = (
    "https://docs.python.org/3/library/logging.config.html#dictionary-schema-details"
)



import logging
import typing
from enum import Enum
from typing import Dict, Text, Any, List, Union, Optional

logger = logging.getLogger(__name__)

if typing.TYPE_CHECKING:  # pragma: no cover
    from rasa_sdk import Tracker
    from rasa_sdk.types import DomainDict


class SlotMapping(Enum):
    """Defines the available slot mappings."""

    FROM_ENTITY = 0
    FROM_INTENT = 1
    FROM_TRIGGER_INTENT = 2
    FROM_TEXT = 3
    CUSTOM = 4

    def __str__(self) -> Text:
        """Returns a string representation of the object."""
        return self.name.lower()

    @staticmethod
    def to_list(x: Optional[Any]) -> List[Any]:
        """Convert object to a list if it isn't."""
        if x is None:
            x = []
        elif not isinstance(x, list):
            x = [x]

        return x

    @staticmethod
    def from_entity(
        entity: Text,
        intent: Optional[Union[Text, List[Text]]] = None,
        not_intent: Optional[Union[Text, List[Text]]] = None,
        role: Optional[Text] = None,
        group: Optional[Text] = None,
    ) -> Dict[Text, Any]:
        """A dictionary for slot mapping to extract slot value.

        From:
        - an extracted entity
        - conditioned on
            - intent if it is not None
            - not_intent if it is not None,
                meaning user intent should not be this intent
            - role if it is not None
            - group if it is not None
        """
        intent, not_intent = (
            SlotMapping.to_list(intent),
            SlotMapping.to_list(not_intent),
        )

        return {
            "type": str(SlotMapping.FROM_ENTITY),
            "entity": entity,
            "intent": intent,
            "not_intent": not_intent,
            "role": role,
            "group": group,
        }

    @staticmethod
    def from_trigger_intent(
        value: Any,
        intent: Optional[Union[Text, List[Text]]] = None,
        not_intent: Optional[Union[Text, List[Text]]] = None,
    ) -> Dict[Text, Any]:
        """A dictionary for slot mapping to extract slot value.

        From:
        - trigger_intent: value pair
        - conditioned on
            - intent if it is not None
            - not_intent if it is not None,
                meaning user intent should not be this intent

        Only used on form activation.
        """
        intent, not_intent = (
            SlotMapping.to_list(intent),
            SlotMapping.to_list(not_intent),
        )

        return {
            "type": str(SlotMapping.FROM_TRIGGER_INTENT),
            "value": value,
            "intent": intent,
            "not_intent": not_intent,
        }

    @staticmethod
    def from_intent(
        value: Any,
        intent: Optional[Union[Text, List[Text]]] = None,
        not_intent: Optional[Union[Text, List[Text]]] = None,
    ) -> Dict[Text, Any]:
        """A dictionary for slot mapping to extract slot value.

        From:
        - intent: value pair
        - conditioned on
            - intent if it is not None
            - not_intent if it is not None,
                meaning user intent should not be this intent
        """
        intent, not_intent = (
            SlotMapping.to_list(intent),
            SlotMapping.to_list(not_intent),
        )

        return {
            "type": str(SlotMapping.FROM_INTENT),
            "value": value,
            "intent": intent,
            "not_intent": not_intent,
        }

    @staticmethod
    def from_text(
        intent: Optional[Union[Text, List[Text]]] = None,
        not_intent: Optional[Union[Text, List[Text]]] = None,
    ) -> Dict[Text, Any]:
        """A dictionary for slot mapping to extract slot value.

        From:
        - a whole message
        - conditioned on
            - intent if it is not None
            - not_intent if it is not None,
                meaning user intent should not be this intent
        """
        intent, not_intent = (
            SlotMapping.to_list(intent),
            SlotMapping.to_list(not_intent),
        )

        return {
            "type": str(SlotMapping.FROM_TEXT),
            "intent": intent,
            "not_intent": not_intent,
        }

    @staticmethod
    def intent_is_desired(
        mapping: Dict[Text, Any],
        tracker: "Tracker",
        domain: "DomainDict",
    ) -> bool:
        """Check whether user intent matches intent conditions."""
        mapping_intents = SlotMapping.to_list(mapping.get("intent", []))
        mapping_not_intents = SlotMapping.to_list(mapping.get("not_intent", []))

        active_loop_name = tracker.active_loop_name
        if active_loop_name:
            mapping_not_intents = (
                mapping_not_intents
                + SlotMapping._get_ignored_intents(mapping, domain, active_loop_name)
            )

        intent = tracker.latest_message.get("intent", {}).get("name")

        intent_not_excluded = not mapping_intents and intent not in mapping_not_intents

        return intent_not_excluded or intent in mapping_intents

    @staticmethod
    def entity_is_desired(
        mapping: Dict[Text, Any],
        tracker: "Tracker",
    ) -> bool:
        """Checks whether slot should be filled by an entity in the input or not.

        Args:
            mapping: Slot mapping.
            tracker: The tracker.

        Returns:
            True, if slot should be filled, false otherwise.
        """
        slot_fulfils_entity_mapping = False
        extracted_entities = tracker.latest_message.get("entities", [])

        for entity in extracted_entities:
            if (
                mapping.get("entity") == entity["entity"]
                and mapping.get("role") == entity.get("role")
                and mapping.get("group") == entity.get("group")
            ):
                matching_values = tracker.get_latest_entity_values(
                    mapping.get("entity", ""),
                    mapping.get("role"),
                    mapping.get("group"),
                )
                slot_fulfils_entity_mapping = matching_values is not None
                break

        return slot_fulfils_entity_mapping

    @staticmethod
    def _get_ignored_intents(
        mapping: Dict[Text, Any],
        domain: "DomainDict",
        active_loop_name: Text,
    ) -> List[Text]:
        mapping_conditions = mapping.get("conditions")
        active_loop_match = False
        ignored_intents = []

        if mapping_conditions:
            for condition in mapping_conditions:
                if condition.get("active_loop") == active_loop_name:
                    active_loop_match = True
                    break

        if active_loop_match:
            form_ignored_intents = domain.get("forms", {})[active_loop_name].get(
                "ignored_intents", []
            )
            if not isinstance(form_ignored_intents, list):
                ignored_intents = [form_ignored_intents]
            else:
                ignored_intents = form_ignored_intents

        return ignored_intents



import typing
from typing import Text, Dict, List, Any

from rasa_sdk import Action
from rasa_sdk.events import SlotSet
from rasa_sdk.knowledge_base.utils import (
    SLOT_OBJECT_TYPE,
    SLOT_LAST_OBJECT_TYPE,
    SLOT_ATTRIBUTE,
    reset_attribute_slots,
    SLOT_MENTION,
    SLOT_LAST_OBJECT,
    SLOT_LISTED_OBJECTS,
    get_object_name,
    get_attribute_slots,
    match_extracted_entities_to_object_type,
)
from rasa_sdk import utils
from rasa_sdk.executor import CollectingDispatcher
from rasa_sdk.interfaces import Tracker
from rasa_sdk.knowledge_base.storage import KnowledgeBase

if typing.TYPE_CHECKING:  # pragma: no cover
    from rasa_sdk.types import DomainDict


class ActionQueryKnowledgeBase(Action):
    """
    Action that queries the knowledge base for objects and attributes of an object.
    The action needs to be inherited and the knowledge base needs to be set.
    In order to actually query the knowledge base you need to:
    - create your knowledge base
    - add mandatory slots to the domain file: 'object_type', 'attribute', 'mention'
    - create NLU data where the required objects are annotated
    - create a story that includes this action
    - add the intent and action to domain file
    """

    def __init__(
        self, knowledge_base: KnowledgeBase, use_last_object_mention: bool = True
    ) -> None:
        self.knowledge_base = knowledge_base
        self.use_last_object_mention = use_last_object_mention

    def name(self) -> Text:
        return "action_query_knowledge_base"

    def utter_attribute_value(
        self,
        dispatcher: CollectingDispatcher,
        object_name: Text,
        attribute_name: Text,
        attribute_value: Text,
    ):
        """
        Utters a response that informs the user about the attribute value of the
        attribute of interest.

        Args:
            dispatcher: the dispatcher
            object_name: the name of the object
            attribute_name: the name of the attribute
            attribute_value: the value of the attribute
        """
        if attribute_value:
            dispatcher.utter_message(
                text=(
                    f"'{object_name}' has the value '{attribute_value}' "
                    f"for attribute '{attribute_name}'."
                )
            )
        else:
            dispatcher.utter_message(
                text=(
                    f"Did not find a valid value for attribute '{attribute_name}' "
                    f"for object '{object_name}'."
                )
            )

    async def utter_objects(
        self,
        dispatcher: CollectingDispatcher,
        object_type: Text,
        objects: List[Dict[Text, Any]],
    ):
        """
        Utters a response to the user that lists all found objects.

        Args:
            dispatcher: the dispatcher
            object_type: the object type
            objects: the list of objects
        """
        if objects:
            dispatcher.utter_message(
                text=f"Found the following objects of type '{object_type}':"
            )

            repr_function = await utils.call_potential_coroutine(
                self.knowledge_base.get_representation_function_of_object(object_type)
            )

            for i, obj in enumerate(objects, 1):
                dispatcher.utter_message(text=f"{i}: {repr_function(obj)}")
        else:
            dispatcher.utter_message(
                text=f"I could not find any objects of type '{object_type}'."
            )

    async def run(
        self,
        dispatcher: CollectingDispatcher,
        tracker: Tracker,
        domain: "DomainDict",
    ) -> List[Dict[Text, Any]]:
        """
        Executes this action. If the user ask a question about an attribute,
        the knowledge base is queried for that attribute. Otherwise, if no
        attribute was detected in the latest request it assumes user is talking
        about a new object type and, multiple objects of the requested type are
        returned from the knowledge base.

        Args:
            dispatcher: the dispatcher
            tracker: the tracker
            domain: the domain

        Returns: list of slots

        """
        object_type = tracker.get_slot(SLOT_OBJECT_TYPE)
        last_object_type = tracker.get_slot(SLOT_LAST_OBJECT_TYPE)
        attribute = tracker.get_slot(SLOT_ATTRIBUTE)
        has_mention = tracker.get_slot(SLOT_MENTION) is not None

        # check if attribute entity is found in latest user message. This is used
        # to track whether the request is to query objects or query attributes
        has_attribute_in_latest_message = any(
            entity.get("entity") == "attribute"
            for entity in tracker.latest_message["entities"]
        )

        if not object_type:
            # sets the object type dynamically from entities if object_type is not
            # found in user query
            object_types = self.knowledge_base.get_object_types()
            object_type = match_extracted_entities_to_object_type(tracker, object_types)
            set_object_type_slot_event = [SlotSet(SLOT_OBJECT_TYPE, object_type)]
            tracker.add_slots(
                set_object_type_slot_event
            )  # temporarily set the `object_type_slot` to extracted value

        if object_type and not has_attribute_in_latest_message:
            return await self._query_objects(dispatcher, object_type, tracker)
        elif object_type and attribute:
            return await self._query_attribute(
                dispatcher, object_type, attribute, tracker
            )

        if last_object_type and has_mention and attribute:
            return await self._query_attribute(
                dispatcher, last_object_type, attribute, tracker
            )

        dispatcher.utter_message(response="utter_ask_rephrase")
        return []

    async def _query_objects(
        self, dispatcher: CollectingDispatcher, object_type: Text, tracker: Tracker
    ) -> List[Dict]:
        """
        Queries the knowledge base for objects of the requested object type and
        outputs those to the user. The objects are filtered by any attribute the
        user mentioned in the request.

        Args:
            dispatcher: the dispatcher
            tracker: the tracker

        Returns: list of slots
        """
        object_attributes = await utils.call_potential_coroutine(
            self.knowledge_base.get_attributes_of_object(object_type)
        )
        # get all set attribute slots of the object type to be able to filter the
        # list of objects
        attributes = get_attribute_slots(tracker, object_attributes)
        # query the knowledge base
        objects = await utils.call_potential_coroutine(
            self.knowledge_base.get_objects(object_type, attributes)
        )
        await utils.call_potential_coroutine(
            self.utter_objects(dispatcher, object_type, objects)
        )

        if not objects:
            return reset_attribute_slots(tracker, object_attributes)

        key_attribute = await utils.call_potential_coroutine(
            self.knowledge_base.get_key_attribute_of_object(object_type)
        )

        last_object = None if len(objects) > 1 else objects[0][key_attribute]

        # To prevent the user to first ask to list the objects for an object type,
        # the object type has to be extracted while the action is executed.
        # Therefore we need to reset the SLOT_OBJECT_TYPE to
        # None to enable this functionality.

        slots = [
            SlotSet(SLOT_OBJECT_TYPE, None),
            SlotSet(SLOT_MENTION, None),
            SlotSet(SLOT_ATTRIBUTE, None),
            SlotSet(SLOT_LAST_OBJECT, last_object),
            SlotSet(SLOT_LAST_OBJECT_TYPE, object_type),
            SlotSet(
                SLOT_LISTED_OBJECTS, list(map(lambda e: e[key_attribute], objects))
            ),
        ]

        return slots + reset_attribute_slots(tracker, object_attributes)

    async def _query_attribute(
        self,
        dispatcher: CollectingDispatcher,
        object_type: Text,
        attribute: Text,
        tracker: Tracker,
    ) -> List[Dict]:
        """
        Queries the knowledge base for the value of the requested attribute of the
        mentioned object and outputs it to the user.

        Args:
            dispatcher: the dispatcher
            tracker: the tracker

        Returns: list of slots
        """
        object_name = get_object_name(
            tracker,
            self.knowledge_base.ordinal_mention_mapping,
            self.use_last_object_mention,
        )

        if object_name is None or not attribute:
            dispatcher.utter_message(response="utter_ask_rephrase")
            return [SlotSet(SLOT_MENTION, None)]

        object_of_interest = await utils.call_potential_coroutine(
            self.knowledge_base.get_object(object_type, object_name)
        )

        if not object_of_interest or attribute not in object_of_interest:
            dispatcher.utter_message(response="utter_ask_rephrase")
            return [SlotSet(SLOT_MENTION, None)]

        value = object_of_interest[attribute]

        object_repr_func = await utils.call_potential_coroutine(
            self.knowledge_base.get_representation_function_of_object(object_type)
        )

        object_representation = object_repr_func(object_of_interest)

        key_attribute = await utils.call_potential_coroutine(
            self.knowledge_base.get_key_attribute_of_object(object_type)
        )

        object_identifier = object_of_interest[key_attribute]

        await utils.call_potential_coroutine(
            self.utter_attribute_value(
                dispatcher, object_representation, attribute, value
            )
        )

        # To prevent the user to first ask to list the objects for an object type,
        # the object type has to be extracted while the action is executed.
        # Therefore we need to reset the SLOT_OBJECT_TYPE to
        # None to enable this functionality.

        slots = [
            SlotSet(SLOT_OBJECT_TYPE, None),
            SlotSet(SLOT_ATTRIBUTE, None),
            SlotSet(SLOT_MENTION, None),
            SlotSet(SLOT_LAST_OBJECT, object_identifier),
            SlotSet(SLOT_LAST_OBJECT_TYPE, object_type),
        ]

        return slots



import logging

import rasa_sdk.cli
import rasa_sdk.version
import rasa_sdk.plugin
from rasa_sdk.interfaces import Tracker, Action, ActionExecutionRejection  # noqa: F401
from rasa_sdk.forms import ValidationAction, FormValidationAction  # noqa: F401

logger = logging.getLogger(__name__)

__version__ = rasa_sdk.version.__version__

if __name__ == "__main__":
    import rasa_sdk.__main__

    rasa_sdk.__main__.main()
